from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _

class ActionsConfig(AppConfig):
    name = 'admin_helpers.actions'
    verbose_name = _('Actions')
    default_auto_field = 'django.db.models.AutoField'




import abc
from typing import List, Dict

from django.db import models


class GlobalSearchMixin:
    objects: models.Manager

    search_order_by = None
    order_in_search = None

    search_icon = 'fa-solid fa-circle'
    search_description = None

    @classmethod
    def get_model_info(cls):
        return {
            'id': cls._meta.model_name,
            'name': cls._meta.verbose_name,
            'icon': cls.search_icon,
            'description': cls.search_description or cls.__doc__ or '-'
        }

    @classmethod
    @abc.abstractmethod
    def get_global_search_fields(cls, query_string: str) -> List[str]:
        ...

    @classmethod
    def get_search_item(cls, obj, request) -> Dict[str, str]:
        return {
            'type': 'model',
            'icon': cls.search_icon,
            'title': str(obj),
            'description': repr(obj),
            'url': None,
            'model_name': cls._meta.verbose_name
        }



from django.conf import settings


IS_GLOBAL_SEARCH_ENABLED = 'admin_helpers.global_search' in settings.INSTALLED_APPS
IS_OBJECT_ACTION_ENABLED = 'admin_helpers.actions' in settings.INSTALLED_APPS



APP_SETTINGS = {
    'IS_GLOBAL_SEARCH_ENABLED': IS_GLOBAL_SEARCH_ENABLED,
    'IS_OBJECT_ACTION_ENABLED': IS_OBJECT_ACTION_ENABLED,
}


from abc import ABC, abstractmethod
from typing import Generator, Type, TypeAlias, List, Dict, Any

from ..mixin import GlobalSearchMixin

SearchResult: TypeAlias = List[Dict[str, Any]]


class BaseSearchBackend(ABC):

    def __init__(self):
        self.request = None

    @abstractmethod
    def get_all_available_models(self, model_name: str | None = None): ...

    @abstractmethod
    def check_perm(self, model, user) -> bool: ...

    @abstractmethod
    def get_available_models(self, request, model_name: str | None = None) -> Generator[Type[GlobalSearchMixin], None, None]: ...

    def handle(self, query_string: str):
        return query_string

    @abstractmethod
    def search(self, request, query_string: str, model: str = None, limit: int = 10) -> SearchResult: ...

    def __call__(self, request, query_string: str, model: str = None, limit: int = 10) -> SearchResult:
        self.request = request
        query_string = self.handle(query_string)
        if not query_string or len(query_string.strip()) < 3:
            return []
        return self.search(request, query_string, model, limit)



from django.contrib.auth.decorators import login_required
from django.http import JsonResponse
from django.views.decorators.csrf import csrf_exempt

from .search_helper import search, get_available_models
import logging

logger = logging.getLogger(__name__)


@login_required
@csrf_exempt
def global_search(request):
    query = request.GET.get('query')
    if not query:
        return JsonResponse({'results': []})

    # Get the model parameter if provided
    model = request.GET.get('model')

    results = search(request, query, model=model)
    return JsonResponse({'results': results})


@login_required
@csrf_exempt
def get_global_search_models(request):
    results = []
    for model in get_available_models(request):
        results.append(model.get_model_info())

    return JsonResponse({'results': results})



import contextlib
import logging
from typing import Generator, Type, Any, Dict

from django.apps import apps
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Model, QuerySet
from django.urls import NoReverseMatch

from .base import BaseSearchBackend, SearchResult
from ..mixin import GlobalSearchMixin

logger = logging.getLogger(__name__)


def get_global_search_options() -> Dict[str, Any]:
    return {
        'exclude': []
    }


class ModelSearchBackend(BaseSearchBackend):

    def get_all_available_models(self, model_name: str | None = None):
        models = apps.get_models(include_auto_created=False)
        available_models = []
        for model in models:
            if model_name and model._meta.model_name != model_name:
                continue
            if issubclass(model, GlobalSearchMixin) and not self.is_exclude_from_search(model):
                available_models.append(model)
        return sorted(available_models, key=lambda x: x.order_in_search or 999)

    def check_perm(self, model, user) -> bool:
        return user.has_perm(f'{model._meta.app_label}.view_{model._meta.model_name}')

    def get_available_models(self, request, model_name: str | None = None) -> Generator[
        Type[GlobalSearchMixin], None, None]:
        for model in self.get_all_available_models(model_name):
            if self.check_perm(model, request.user):  # noqa
                yield model  # noqa

    def search(self, request, query_string: str, model: str = None, limit: int = 10) -> SearchResult:
        results = []

        for m in self.get_available_models(request, model):
            try:
                results.extend(self.search_results(m, query_string, limit))
            except Exception as e:
                logger.error(f"Error searching model {m._meta.model_name}: {e}")

        return results

    @staticmethod
    def is_exclude_from_search(model: Type[GlobalSearchMixin]) -> bool:
        options = get_global_search_options()
        exclude_list = list(map(lambda x: str(x).lower(), options.get('exclude', [])))
        full_model_name = f'{model._meta.app_label}.{model._meta.model_name}'
        return (full_model_name in exclude_list) or (model._meta.app_label in exclude_list)

    @classmethod
    def get_global_search_results(cls, model: Type[GlobalSearchMixin], query_string: str) -> QuerySet[Any]:
        query = cls.get_query_orm(model, query_string)
        if query is None:
            return model.objects.none()
        qs = model.objects.using('default')

        with contextlib.suppress(ValidationError):
            qs = qs.filter(query)
            if model.search_order_by:
                qs = qs.order_by(model.search_order_by)
            return qs.distinct()
        return model.objects.none()

    def search_results(self, model: Type[GlobalSearchMixin], query_string: str, limit: int = 10):
        for obj in self.get_global_search_results(model, query_string)[:limit]:
            item = model.get_search_item(obj, self.request)
            if item.get('url') is None:
                item['url'] = self.get_item_url(obj)
            yield item

    @classmethod
    def _get_field_modificator(cls, field_name: str) -> str:
        def get_field_name(name: str, char: str) -> str | None:
            if name.startswith(char):
                return name[1:]
            return None

        if name := get_field_name(field_name, '='):
            return f'{name}__iexact'
        if name := get_field_name(field_name, '~'):
            return f'{name}__icontains'
        return field_name

    @classmethod
    def get_query_orm(cls, model: Type[GlobalSearchMixin], query_string: str):
        query = None
        if not query_string:
            return query
        with contextlib.suppress(ValidationError):
            for field in model.get_global_search_fields(query_string):
                field_modificator = cls._get_field_modificator(field)
                if query is None:
                    query = models.Q(**{field_modificator: query_string})
                else:
                    query |= models.Q(**{field_modificator: query_string})

        return query

    @staticmethod
    def get_item_url(obj: GlobalSearchMixin) -> str:
        try:
            from django.urls import reverse
            _meta = obj._meta
            app_label = _meta.app_label
            model_name = _meta.model_name
            return reverse(f'admin:{app_label}_{model_name}_change', args=(obj.pk,))
        except NoReverseMatch:
            return '#'



from datetime import timedelta
from typing import Literal

from django import template
from django.urls import reverse, NoReverseMatch
from django.utils.dateparse import parse_datetime

from admin_helpers.utils import pluralize, render_badge

register = template.Library()


@register.filter(name='qs_order')
def query_set_order(qs, ordering):
    try:
        ordering = ordering.split(',')
        return qs.order_by(*ordering)
    except Exception:
        return qs


@register.inclusion_tag('admin_helpers/copy_inline_btn.html', name='copy_inline_btn')
def insert_copy_inline_button(value, copy_value=None):
    return {'value': value, 'copy_value': copy_value}


@register.filter(name='repr_object')
def repr_object(obj):
    return repr(obj)


@register.filter(name='parse_datetime')
def parse_date_value(value):
    try:
        return parse_datetime(value)
    except Exception:
        return ''


@register.filter(name='str_split')
def split_string(value: str, sep=' '):
    return value.split(sep)


@register.filter(name='str_rsplit')
def rsplit_string(value: str, sep=' '):
    return value.rsplit(sep)


@register.simple_tag(name='url_safe')
def safe_url_reverse(view_name, *args, **kwargs):
    try:
        return reverse(view_name, args=args, kwargs=kwargs)
    except NoReverseMatch:
        return '#'


@register.filter
def format_time(value):
    """
    Форматирование времени из объекта timedelta в человеко-читаемом формате.
    Пример: "1 минута 51 секунда".
    """
    if not isinstance(value, timedelta):
        try:
            value = timedelta(seconds=int(value))
        except (ValueError, TypeError):
            return value

    total_seconds = int(value.total_seconds())
    hours = total_seconds // 3600
    minutes = (total_seconds % 3600) // 60
    seconds = total_seconds % 60

    # Формируем человеко-читаемое время
    parts = []
    if hours > 0:
        parts.append(f'{hours} {pluralize(hours, ('час', 'часа', 'часов'))}')
    if minutes > 0:
        parts.append(f'{minutes} {pluralize(minutes, ('минута', 'минуты', 'минут'))}')
    if seconds > 0 or (hours == 0 and minutes == 0):
        parts.append(f'{seconds} {pluralize(seconds, ('секунда', 'секунды', 'секунд'))}')

    return " ".join(parts)


@register.inclusion_tag('admin_helpers/badge.html', name='badge')
def insert_badge(
        value: str,
        level: Literal['info', 'success', 'warning', 'warning'] = 'info',
        large: bool = False,
        class_list: list[str] = None
):
    return render_badge(value, level, large, class_list=class_list, return_ctx=True)



from typing import Type, List

from django.apps import apps

from .backends.base import SearchResult
from .mixin import GlobalSearchMixin


def switch_layout(value: str) -> str:
    # Сопоставление символов русской и английской раскладки
    ru = 'ёйцукенгшщзхъфывапролджэячсмитьбю.'
    en = '`qwertyuiop[]asdfghjkl;\'zxcvbnm,./'
    table = {}

    # Рус -> Англ
    for r, e in zip(ru, en):
        table[r] = e
        table[r.upper()] = e.upper()
    # Англ -> Рус
    for r, e in zip(ru, en):
        table[e] = r
        table[e.upper()] = r.upper()

    # Перекладываем строку
    return ''.join([table.get(char, char) for char in value])


def search_pages(request, query: str):
    from django.contrib import admin
    pages = []
    switched_query = switch_layout(query).strip().lower()

    for app in admin.site.get_app_list(request):
        if not app['has_module_perms']:
            continue
        for model in app['models']:
            name = model['name']
            if query.strip().lower() not in name.lower() and switched_query not in name.lower():
                continue
            model_cls = model['model']
            pages.append({
                'type': 'page',
                'icon': getattr(model_cls, 'search_icon', 'fa-solid fa-square'),
                'title': name,
                'description': '',
                'url': model['admin_url'],
                'model_name': str(app.get('name', app['app_label'])).upper(),
            })
    return pages


def search(request, query: str, model: str = None, limit: int = 10) -> SearchResult:
    app_conf = apps.get_app_config('global_search')
    backend_class = app_conf.search_backend
    results = [
        *search_pages(request, query),
    ]
    if backend_class:
        backend = backend_class()
        results.extend(backend(request, query, model=model, limit=limit))
    return results

def get_available_models(request) -> List[Type[GlobalSearchMixin]]:
    app_conf = apps.get_app_config('global_search')
    backend_class = app_conf.search_backend
    if backend_class:
        backend = backend_class()
        return list(backend.get_available_models(request))
    return []


import asyncio
import contextlib
import math
import time
from functools import partial
from typing import Literal

from django.db import connection
from django.template.loader import render_to_string


def pluralize(number, titles):
    """
    Возвращает форму слова в зависимости от числа.
    :param number: Число, для которого определяется форма слова
    :param titles: Список из трех форм слова: для 1, для 2, для 5
    :return: Строка с правильной формой слова
    """
    cases = [2, 0, 1, 1, 1, 2]
    if 4 < number % 100 < 20:
        index = 2
    else:
        index = cases[min(number % 10, 5)]
    return titles[index]


pluralize_days = partial(pluralize, titles=('день', 'дня', 'дней'))


def convert_size(size_bytes):  # pragma: no cover
    """Конвертирует размер"""
    if size_bytes == 0:
        return "0 B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return "%s %s" % (s, size_name[i])


@contextlib.contextmanager
def suppress_autotime(model, fields):  # pragma: no cover
    _original_values = {}
    for field in model._meta.local_fields:
        if field.name in fields:
            _original_values[field.name] = {
                'auto_now': field.auto_now,
                'auto_now_add': field.auto_now_add,
            }
            field.auto_now = False
            field.auto_now_add = False
    try:
        yield
    finally:
        for field in model._meta.local_fields:
            if field.name in fields:
                field.auto_now = _original_values[field.name]['auto_now']
                field.auto_now_add = _original_values[field.name]['auto_now_add']


def execute_raw_sql(sql_query):
    with connection.cursor() as cursor:
        cursor.execute(sql_query)
        columns = [col[0] for col in cursor.description]
        results = [
            dict(zip(columns, row))
            for row in cursor.fetchall()
        ]
    return results


def measure_time(func):
    if asyncio.iscoroutinefunction(func):
        async def wrapper(*args, **kwargs):
            start_time = time.time()  # Записываем время начала
            result = await func(*args, **kwargs)
            end_time = time.time()  # Записываем время окончания
            elapsed_time = end_time - start_time
            print(f"Функция '{func.__name__}' выполнилась за {elapsed_time:.6f} секунд.")
            return result
    else:
        def wrapper(*args, **kwargs):
            start_time = time.time()  # Записываем время начала
            result = func(*args, **kwargs)
            end_time = time.time()  # Записываем время окончания
            elapsed_time = end_time - start_time
            print(f"Функция '{func.__name__}' выполнилась за {elapsed_time:.6f} секунд.")
            return result
    return wrapper


def is_valid_uuid(value: str) -> bool:
    try:
        from uuid import UUID
        UUID(value)
        return True
    except ValueError:
        return False


def is_has_kwargs(func) -> bool:
    import inspect
    sig = inspect.signature(func)
    return any(sig.kind == inspect.Parameter.VAR_KEYWORD for sig in sig.parameters.values())


def uuid_to_color(uuid_str: str, as_rgba: bool = False, opacity: float = 1.0) -> str:
    """
    Преобразует UUID в цвет в формате HEX (#RRGGBB).

    Функция принимает строку UUID и возвращает соответствующий цвет в формате HEX.
    Это полезно для визуального различения объектов на основе их UUID.

    Args:
        uuid_str (str): Строка UUID для преобразования в цвет.
        as_rgba (bool):
        opacity (float):

    Returns:
        str: Цвет в формате HEX (#RRGGBB).
    """
    # Проверяем, является ли строка валидным UUID
    if not is_valid_uuid(uuid_str):
        # Если строка не является валидным UUID, возвращаем дефолтный цвет
        return "#000000"

    # Импортируем UUID для работы с UUID
    from uuid import UUID

    # Преобразуем строку в объект UUID
    uuid_obj = UUID(uuid_str)

    # Получаем байты UUID и берем первые 3 байта для RGB
    uuid_bytes = uuid_obj.bytes

    # Используем первые 3 байта для создания цвета
    r = uuid_bytes[0]
    g = uuid_bytes[1]
    b = uuid_bytes[2]
    if as_rgba:
        return f"rgba({r}, {g}, {b}, {opacity})"
    # Формируем HEX-строку цвета
    color_hex = f"#{r:02x}{g:02x}{b:02x}"

    return color_hex


def render_badge(
        value: str,
        level: Literal['info', 'success', 'warning', 'warning'] = 'info',
        large: bool = False,
        class_list: list[str] = None,
        return_ctx: bool = False,
):
    class_list = class_list or []
    if large:
        class_list.append('large')
    ctx = {
        'value': value,
        'level': level,
        'class_list': class_list
    }
    if return_ctx:
        return ctx
    return render_to_string('admin_helpers/badge.html', ctx)


from django.apps import AppConfig
from django.conf import settings
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _


class GlobalSearchConfig(AppConfig):
    name = 'admin_helpers.global_search'
    verbose_name = _('Global Search')
    default_auto_field = 'django.db.models.AutoField'

    @property
    def search_backend(self):
        backend_class = getattr(
            settings,
            'DAH_GLOBAL_SEARCH_BACKEND',
            'admin_helpers.global_search.backends.model.ModelSearchBackend'
        )

        try:
            backend = import_string(backend_class)
        except ImportError:
            return None

        return backend



from django.urls import path

from .views import global_search, get_global_search_models

app_name = 'global_search'
urlpatterns = [
    path('search', global_search, name='global_search'),
    path('get-search-models', get_global_search_models, name='get_global_search_models'),
]
